#include "opencv2/opencv.hpp"

void printDeltaMilliSec(double tick_begin, double tick_end)
{
    double delta = 1000.0 * (tick_end - tick_begin) / cv::getTickFrequency();
    printf("%f [ms]\n", delta);
}

void onMouse(int event, int x, int y, int flags, void *param)
{
    if (event == CV_EVENT_LBUTTONDOWN) {
        cv::Point *tcenterp = (cv::Point *)param;
        tcenterp->x = x;
        tcenterp->y = y;
    }
}

int SSD(const cv::Mat& Iblock, const cv::Mat& Jblock)
{
    int ssd = 0;
    for (int j = 0; j < Iblock.rows; j++) {
        for (int i = 0; i < Iblock.cols; i++) {
            int d = Iblock.at<uchar>(j, i) - Jblock.at<uchar>(j, i);
            ssd += d * d;
        }
    }
    return ssd;
}

void trackBM(const cv::Mat& J, const cv::Mat& Iblock, cv::Point& tcenter,
             cv::Size search_size)
{
    double tick_begin = (double)cv::getTickCount();

    int sxmin = tcenter.x - search_size.width / 2;
    int sxmax = tcenter.x + search_size.width / 2;
    int symin = tcenter.y - search_size.height / 2;
    int symax = tcenter.y + search_size.height / 2;

    if (sxmin < 0) { sxmin = 0; }
    if (sxmax >= J.cols - Iblock.cols) { sxmax = J.cols - Iblock.cols - 1; }
    if (symin < 0) { symin = 0; }
    if (symax >= J.rows - Iblock.rows) { symax = J.rows - Iblock.rows - 1; }

    int minssd = INT_MAX;
    cv::Point minloc = cv::Point(0, 0);
    for (int j = symin; j <= symax; j++) {
        for (int i = sxmin; i <= sxmax; i++) {
            cv::Mat Jblock = J(cv::Rect(i, j, Iblock.cols, Iblock.rows));
            int ssd = SSD(Iblock, Jblock);
            if (ssd < minssd) {
                minssd = ssd;
                minloc.x = i;
                minloc.y = j;
            }
        }
    }
    tcenter.x = minloc.x + Iblock.cols / 2;
    tcenter.y = minloc.y + Iblock.rows / 2;

    double tick_end = (double)cv::getTickCount();

    printDeltaMilliSec(tick_begin, tick_end);
}

int main() {
    cv::namedWindow("output", cv::WINDOW_AUTOSIZE);
    cv::VideoCapture cap(0);
    cv::Mat input, output;

    const int bwidth = 64;
    const int bheight = 64;

    cv::Point tcenter; // center of tracked region
    cap >> input;
    tcenter.x = input.cols / 2;
    tcenter.y = input.rows / 2;

    cv::setMouseCallback("output", onMouse, (void *)&tcenter);

    int search_margin = bwidth / 2;
    cv::createTrackbar("search margin", "output", &search_margin, bwidth);

    cv::Mat I, J, Iblock;

    while (1) {
        cap >> input;
        cv::cvtColor(input, J, CV_BGR2GRAY);
        if (I.empty()) { // in case of the first frame
            J.copyTo(I);
        }

        Iblock = I(cv::Rect(tcenter.x - bwidth / 2,
                            tcenter.y - bheight / 2,
                            bwidth, bheight));
        trackBM(J, Iblock, tcenter, cv::Size(bwidth + 2 * search_margin,
                                             bheight + 2 * search_margin));
        J.copyTo(I);

        input.copyTo(output);
        cv::rectangle(output,
                      cv::Point((int)tcenter.x - bwidth / 2,
                                (int)tcenter.y - bheight / 2),
                      cv::Point((int)tcenter.x + bwidth / 2,
                                (int)tcenter.y + bheight / 2), 
                      CV_RGB(255, 0, 0), 3);
        cv::rectangle(output,
                      cv::Point((int)tcenter.x - bwidth / 2 - search_margin,
                                (int)tcenter.y - bheight / 2 - search_margin),
                      cv::Point((int)tcenter.x + bwidth / 2 + search_margin,
                                (int)tcenter.y + bheight / 2 + search_margin), 
                      CV_RGB(0, 255, 0), 1);
        cv::imshow("output", output);
        if (cv::waitKey(30) > 0) {
            break;
        }
    }

    return 0;
}
