#include "opencv2/opencv.hpp"

void printDeltaMilliSec(double tick_begin, double tick_end)
{
    double delta = 1000.0 * (tick_end - tick_begin) / cv::getTickFrequency();
    printf("%f [ms]\n", delta);
}

void onMouse(int event, int x, int y, int flags, void *param)
{
    if (event == CV_EVENT_LBUTTONDOWN) {
        cv::Point2f *tcenterp = (cv::Point2f *)param;
        tcenterp->x = (float)x;
        tcenterp->y = (float)y;
    }
}

void pushImage(std::vector<cv::Mat>& matvec, const cv::Mat& image,
                   int scale, int offset, const std::string& text)
{
    cv::Mat name = cv::Mat::zeros(20, image.cols, CV_16S);
    name -= 32768;
    cv::putText(name, text, cv::Point(1, 16),
                CV_FONT_HERSHEY_PLAIN, 1.0, cvScalar(32767));
    cv::Mat named_image;
    std::vector<cv::Mat> v;
    v.push_back(image * scale + offset);
    v.push_back(name);
    cv::vconcat(&v[0], 2, named_image);

    cv::Mat sep = cv::Mat::zeros(named_image.rows, 4, CV_16S);
    sep -= 32768;
    matvec.push_back(named_image);
    matvec.push_back(sep);
}

void trackLK(const cv::Mat& J, const cv::Mat& Iblock, cv::Point2f& tcenter,
             cv::Mat& illust)
{
    double tick_begin = (double)cv::getTickCount();

    const int max_iter = 20;

    int h = Iblock.rows;
    int w = Iblock.cols;
    cv::Mat Ix(h, w, CV_16S);
    cv::Mat Iy(h, w, CV_16S);
    cv::Mat IxIx(h, w, CV_16S);
    cv::Mat IyIy(h, w, CV_16S);
    cv::Mat IxIy(h, w, CV_16S);

    float g11 = 0.0f, g12 = 0.0f, g22 = 0.0f;
    for (int j = 1; j < h - 1; j++) {
        for (int i = 1; i < w - 1; i++) {
            Ix.at<short>(j, i) =
                (Iblock.at<short>(j, i + 1) - Iblock.at<short>(j, i - 1)) / 2;
            Iy.at<short>(j, i) =
                (Iblock.at<short>(j + 1, i) - Iblock.at<short>(j - 1, i)) / 2;
            IxIx.at<short>(j, i) = Ix.at<short>(j, i) * Ix.at<short>(j, i);
            IxIy.at<short>(j, i) = Ix.at<short>(j, i) * Iy.at<short>(j, i);
            IyIy.at<short>(j, i) = Iy.at<short>(j, i) * Iy.at<short>(j, i);

            g11 += IxIx.at<short>(j, i);
            g12 += IxIy.at<short>(j, i);
            g22 += IyIy.at<short>(j, i);
        }
    }
    cv::Mat G = (cv::Mat_<float>(2, 2) << g11, g12, g12, g22);

    bool is_converged = false;
    cv::Mat Jblock;
    cv::Mat IminJ(h, w, CV_16S);
    cv::Mat IminJ_Ix(h, w, CV_16S);
    cv::Mat IminJ_Iy(h, w, CV_16S);

    int iter;
    for (iter = 0; iter < max_iter && !is_converged; iter++) {
        cv::getRectSubPix(J, Iblock.size(), tcenter, Jblock);
        Jblock.convertTo(Jblock, CV_16S);

        float e1 = 0.0f, e2 = 0.0;
        for (int j = 1; j < h - 1; j++) {
            for (int i = 1; i < w - 1; i++) {
                IminJ.at<short>(j, i) =
                    Iblock.at<short>(j, i) - Jblock.at<short>(j, i);
                IminJ_Ix.at<short>(j, i) =
                    IminJ.at<short>(j, i) * Ix.at<short>(j, i);
                IminJ_Iy.at<short>(j, i) =
                    IminJ.at<short>(j, i) * Iy.at<short>(j, i);

                e1 += IminJ_Ix.at<short>(j, i);
                e2 += IminJ_Iy.at<short>(j, i);
            }
        }

        cv::Mat e = (cv::Mat_<float>(2, 1) << e1, e2);
        cv::Mat sol;
        cv::solve(G, e, sol, cv::DECOMP_SVD);
        
        float dx = sol.at<float>(0);
        float dy = sol.at<float>(1);
        tcenter.x += dx;
        tcenter.y += dy;

        is_converged = (dx * dx + dy * dy < 0.05) ? true : false;
    }

    double tick_end = (double)cv::getTickCount();
    printf("iter = %d\n", iter);
    printDeltaMilliSec(tick_begin, tick_end);


    std::vector<cv::Mat> matvec;
    pushImage(matvec, Iblock, 256, -32768, "I");
    pushImage(matvec, Jblock, 256, -32768, "J");
    pushImage(matvec, IminJ, 256, 0, "I-J");
    pushImage(matvec, Ix, 256, 0, "Ix");
    pushImage(matvec, Iy, 256, 0, "Iy");
    pushImage(matvec, IxIx, 16, 0, "IxIx");
    pushImage(matvec, IxIy, 16, 0, "IxIy");
    pushImage(matvec, IyIy, 16, 0, "IyIy");
    pushImage(matvec, IminJ_Ix, 16, 0, "(I-J)Ix");
    pushImage(matvec, IminJ_Iy, 16, 0, "(I-J)Iy");
    cv::hconcat(&matvec[0], matvec.size(), illust);
}

int main() {
    cv::namedWindow("output", cv::WINDOW_AUTOSIZE);
    cv::namedWindow("illustration", cv::WINDOW_AUTOSIZE);
    cv::VideoCapture cap(0);
    cv::Mat input, output, illust;

    const int bwidth = 64;
    const int bheight = 64;

    cv::Point2f tcenter; // center of tracked region
    cap >> input;
    tcenter.x = (float)input.cols / 2;
    tcenter.y = (float)input.rows / 2;

    cv::setMouseCallback("output", onMouse, (void *)&tcenter);

    cv::Mat I, J, Iblock;

    while (1) {
        cap >> input;
        cv::cvtColor(input, J, CV_BGR2GRAY);
        if (I.empty()) { // in case of the first frame
            J.copyTo(I);
        }

        cv::getRectSubPix(I, cv::Size(bwidth, bheight), tcenter, Iblock);
        Iblock.convertTo(Iblock, CV_16S);
        trackLK(J, Iblock, tcenter, illust);
        J.copyTo(I);

        input.copyTo(output);
        cv::rectangle(output,
                      cv::Point((int)tcenter.x - bwidth / 2,
                                (int)tcenter.y - bheight / 2),
                      cv::Point((int)tcenter.x + bwidth / 2,
                                (int)tcenter.y + bheight / 2), 
                      CV_RGB(255, 0, 0), 3);
        cv::imshow("output", output);
        cv::imshow("illustration", illust);
        if (cv::waitKey(30) > 0) {
            break;
        }
    }

    return 0;
}
